﻿using System;

namespace mosquitto.net
{
    /// <summary>
    /// Interface to the mosquitto client
    /// </summary>
    public interface IMosquitto : IDisposable
    {
        /// <summary>
        /// Library version
        /// </summary>
        string Version { get; }
        /// <summary>
        /// Broker connection callback
        /// </summary>
        event Action<IMosquitto, ConnectionResult> OnConnection;
        /// <summary>
        /// Broker disconnect callback
        /// </summary>
        event Action<IMosquitto, ReturnCode> OnDisconnect;
        
        event Action<IMosquitto, ReturnCode> OnError;

        event Action<IMosquitto, LogPriority,string> OnLog;
        /// <summary>
        /// Initialize the client logging
        /// </summary>
        /// <returns></returns>
        ReturnCode InitialiseLog();
        /// <summary>
        /// Set Will for this client
        /// </summary>
        /// <param name="topic">the topic on which to publish the will</param>
        /// <param name="payloadlen">the size of the payload (bytes)</param>
        /// <param name="payload">data to send</param>
        /// <param name="qos">the Quality of Service to be used for the will</param>
        /// <param name="retain">set to true to make the will a retained message</param>
        /// <returns></returns>
        ReturnCode SetWill( string topic, uint payloadlen, byte[] payload, Qos qos, bool retain);
        /// <summary>
        /// Configure username and password for a mosquitton instance. This is only
        /// supported by brokers that implement the MQTT spec v3.1. By default, no
        /// username or password will be sent.
        /// If username is NULL, the password argument is ignored.
        /// This must be called before calling mosquitto_connect
        /// </summary>
        /// <param name="username">the username to send as a string, or NULL to disable authentication</param>
        /// <param name="password">the password to send as a string. Set to NULL when username is valid in order to send just a username</param>
        /// <returns></returns>
        ReturnCode SetUserNamePassword(string username, string password);
        /// <summary>
        /// Connect to an MQTT broker
        /// </summary>
        /// <param name="host">the hostname or ip address of the broker to connect to</param>
        /// <param name="port">the network port to connect to. Usually 1883</param>
        /// <param name="keepalive">the number of seconds after which the broker should send a PING  message to the client if no other messages have been exchanged
        ///  in that time</param>
        /// <returns></returns>
        ReturnCode Connect(string host, int port, int keepalive);
        /// <summary>
        /// Connect to an MQTT broker asynchonously
        /// </summary>
        /// <param name="host">the hostname or ip address of the broker to connect to</param>
        /// <param name="port">the network port to connect to. Usually 1883</param>
        /// <param name="keepalive">the number of seconds after which the broker should send a PING  message to the client if no other messages have been exchanged
        ///  in that time</param>
        /// <param name="callback">callback on connect</param>
        void ConnectAsync(string host, int port, int keepalive, Action<ReturnCode> callback);
        /// <summary>
        /// Start the client polling thread
        /// </summary>
        void Start();
        /// <summary>
        /// This function provides an easy way of reconnecting to a broker after a
        /// connection has been lost
        /// </summary>
        /// <returns></returns>
        ReturnCode Reconnect();
        /// <summary>
        /// Disconnect from the broker
        /// </summary>
        /// <returns></returns>
        ReturnCode Disconnect();
        /// <summary>
        /// event on a successful subscription
        /// </summary>
        event Action<IMosquitto, int, byte[]> OnSubscribe;
        /// <summary>
        /// event on a message recieved
        /// </summary>
        event Action<IMosquitto, Message> OnMessage;
        /// <summary>
        /// event on unsubscribtion to a message
        /// </summary>
        event Action<IMosquitto, int> OnUnsubscribe;
        /// <summary>
        /// Subscribe to a topic
        /// </summary>
        /// <param name="mid"> a pointer to a uint16_t. If not NULL, the function will set this to
        ///	       the message id of this particular message. This can be then used
        ///	       with the subscribe callback to determine when the message has been
        ///	       sent
        /// </param>
        /// <param name="sub">the subscription pattern</param>
        /// <param name="qos">the requested Quality of Service for this subscription</param>
        /// <returns></returns>
        ReturnCode Subscribe(ref int mid, string sub, Qos qos);
        /// <summary>
        /// Unsubscribe from a topic
        /// </summary>
        /// <param name="mid">a pointer to a uint16_t. If not NULL, the function will set this to
        ///	       the message id of this particular message. This can be then used
        ///	       with the unsubscribe callback to determine when the message has been
        ///	       sent.
        /// </param>
        /// <param name="sub">the unsubscription pattern</param>
        /// <returns></returns>
        ReturnCode Unsubscribe(ref int mid, string sub);
        /// <summary>
        /// Subscribe, return an IObservable
        /// </summary>
        /// <param name="topic">the topic to subscribe to</param>
        /// <param name="qos">Quality required</param>
        /// <returns>an observable</returns>
        IObservable<Message> Subscribe(string topic, Qos qos);
        /// <summary>
        /// Subscribe, return a  subscribtion to an IObservable
        /// </summary>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="qos">Quality required</param>
        /// <param name="subscribe">Action to perform on receipt of message</param>
        /// <returns>IDisposable</returns>
        IDisposable Subscribe(string topic, Qos qos, Action<Message> subscribe);
        /// <summary>
        /// Subscribe, return a  subscribtion to an IObservable
        /// </summary>
        /// <param name="topic">The topic to subscribe to</param>
        /// <param name="qos">Quality required</param>
        /// <param name="subscribe">Action to perform on receipt of message</param>
        /// <param name="where">A filter to apply to the observable</param>
        /// <returns>IDisposable</returns>
        IDisposable Subscribe(string topic, Qos qos, Action<Message> subscribe, Func<Message, bool> where);
        
        /// <summary>
        /// Callback on message published
        /// </summary>
        event Action<IMosquitto, ushort> OnPublished;
        /// <summary>
        /// Publish a message on a given topic
        /// </summary>
        /// <param name="mid">pointer to a uint16_t. If not NULL, the function will set this
        ///  to the message id of this particular message. This can be then
        ///  used with the publish callback to determine when the message
        ///  has been sent.
        ///  Note that although the MQTT protocol doesn't use message ids
        ///  for messages with QoS=0, libmosquitto assigns them message ids
        ///  so they can be tracked with this parameter
        /// </param>
        /// <param name="topic">The topic to publish</param>
        /// <param name="payload">data to send</param>
        /// <param name="qos"> Quality of Service to be used for the message</param>
        /// <param name="retain">set to true to make the message retained</param>
        /// <returns></returns>
        ReturnCode Publish(ref ushort mid, string topic, byte[] payload, Qos qos, bool retain);
    }
}