﻿using System;
using System.Collections.Generic;
using System.Reactive.Linq;
using System.Runtime.InteropServices;
using System.Threading;
using System.Threading.Tasks;

namespace mosquitto.net
{
    public class Mosquitto : IMosquitto
    {
        static Mosquitto()
        {
            libmosquitto.mosquitto_lib_init();

        }

        public event Action<IMosquitto, int, byte[]> OnSubscribe = (mqtt, mid, qos) => { };
        public event Action<IMosquitto, Message> OnMessage = (mqtt,message)=> { };
        public event Action<IMosquitto, int> OnUnsubscribe = (mqtt,mid)=> { };
        public event Action<IMosquitto, LogPriority, string> OnLog = (mqtt,p,msg)=> { };
        public Mosquitto(string id,bool clean)
        {
            _libraryRefCounter++;
            _this = GCHandle.Alloc(this, GCHandleType.Normal);
            IntPtr obj = GCHandle.ToIntPtr(_this);
            _handle = libmosquitto.mosquitto_new(id, clean, obj);

            libmosquitto.mosquitto_connect_callback_set(_handle, ConnectCallback);
            libmosquitto.mosquitto_disconnect_callback_set(_handle, DisconnectCallback);
            libmosquitto.mosquitto_unsubscribe_callback_set(_handle, UnsubscribeCallback);
            libmosquitto.mosquitto_publish_callback_set(_handle, PublishCallback);
            libmosquitto.mosquitto_subscribe_callback_set(_handle, SubscribeCallback);
            libmosquitto.mosquitto_message_callback_set(_handle, MessageCallback);
            libmosquitto.mosquitto_log_init(_handle, LogCallback);
            _token = new CancellationTokenSource();
            _task = new Task(Loop, _token.Token, TaskCreationOptions.LongRunning);

        }

        private void Loop()
        {
            while (!_token.IsCancellationRequested)
            {
                _token.Token.WaitHandle.WaitOne(100);
                Poll();
            }
        }

        private void Poll()
        {
            ReturnCode rc = libmosquitto.mosquitto_loop(_handle,-1, 1000);
            if (rc != ReturnCode.Code.Success && rc != ReturnCode.Code.NoConnection)
            {
                Console.WriteLine("mqtt error - {0}", rc);
                if (Reconnect() != ReturnCode.Code.Success)
                {
                    OnError(this, rc);
                }
                else
                {
                    Resubscribe();
                }
            }
        }

        #region IMosquitto Members

        public void Start()
        {
            Stop();
            _token = new CancellationTokenSource();
            _task = Task.Factory.StartNew(Loop, TaskCreationOptions.LongRunning);
        }

        public event Action<IMosquitto, ConnectionResult> OnConnection = (c, rc) => { };

        public event Action<IMosquitto, ReturnCode> OnDisconnect = (c, rc) => { };

        public event Action<IMosquitto, ReturnCode> OnError = (c, rc) => { };

        public string Version
        {
            get
            {
                int major;
                int minor;
                int revision;
                libmosquitto.mosquitto_lib_version(out major, out minor, out revision);
                return string.Format("{0}.{1}.{2}", major, minor, revision);
            }
        }

        public ReturnCode InitialiseLog()
        {
            return libmosquitto.mosquitto_log_init(_handle,LogCallback);
        }

        public ReturnCode SetWill( string topic, uint payloadlen, byte[] payload, Qos qos, bool retain)
        {
            return libmosquitto.mosquitto_will_set(_handle, topic, payloadlen, payload,(int) qos, retain);
        }

        public ReturnCode SetUserNamePassword(string username, string password)
        {
            return libmosquitto.mosquitto_username_pw_set(_handle, username, password);
        }

        public ReturnCode Connect(string host, int port, int keepalive)
        {
            return libmosquitto.mosquitto_connect(_handle, host, port, keepalive);
        }

        public void ConnectAsync(string host, int port, int keepalive, Action<ReturnCode> callback)
        {
            Task<ReturnCode>.Factory.StartNew(() => Connect(host, port, keepalive)).ContinueWith(t => callback(t.Result));
        }

        public ReturnCode Reconnect()
        {
            return libmosquitto.mosquitto_reconnect(_handle);
        }

        public ReturnCode Disconnect()
        {
            return libmosquitto.mosquitto_disconnect(_handle);
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        public event Action<IMosquitto, ushort> OnPublished = (mqtt, mid) => { };

        public ReturnCode Publish(ref ushort mid, string topic, byte[] payload, Qos qos, bool retain)
        {
            return libmosquitto.mosquitto_publish(_handle, ref mid, topic, (uint) payload.Length, payload, (int) qos,
                                                  retain);
        }

        public ReturnCode Subscribe(ref int mid, string topic, Qos qos)
        {
            ReturnCode rc = libmosquitto.mosquitto_subscribe(_handle, ref mid, topic, (int)qos);
            if(!_subscriptions.ContainsKey(topic))
            {
                _subscriptions.Add(topic,qos);
            }
            return rc;
        }

        public ReturnCode Unsubscribe(ref int mid, string sub)
        {
            ReturnCode rc = libmosquitto.mosquitto_unsubscribe(_handle, ref mid, sub);
            _subscriptions.Remove(sub);
            return rc;
        }

        public IDisposable Subscribe(string topic, Qos qos, Action<Message> subscribe,Func<Message,bool> where )
        {
            var observable = Subscribe(topic, qos).Where(where);
            return observable.Subscribe(subscribe);
        }
        
        public IDisposable Subscribe(string topic, Qos qos,Action<Message> subscribe)
        {
            var observable = Subscribe(topic, qos);
            return observable.Subscribe(subscribe);
        }

        public IObservable<Message> Subscribe(string topic, Qos qos)
        {
            int mid = 0;
            var result = Subscribe(ref mid, topic, qos);
            if (result == ReturnCode.Code.Success)
            {
                return _observable ?? (_observable = Observable.Create<Message>(o =>
                                                                                    {
                                                                                        Action<IMosquitto, Message> handler = (mosquitto, m) => o.OnNext(m);
                                                                                        Action<IMosquitto, ReturnCode> disconnect = (mosq, rc) => o.OnCompleted();
                                                                                        Action<IMosquitto, ReturnCode> error = (mosq, rc) => o.OnError(new MosquittoException(rc));
                                                                                        OnError += error;
                                                                                        OnDisconnect += disconnect;
                                                                                        OnMessage += handler;
                                                                                        return () =>
                                                                                                   {
                                                                                                       OnMessage -= handler;
                                                                                                       OnError -= error;
                                                                                                       OnDisconnect -= disconnect;
                                                                                                   };
                                                                                    }));
            }
            throw new MosquittoException(result);
        }

        
        #endregion

        private static void OnConnectCallback(IntPtr mosquitto, IntPtr cookie, int rc)
        {
            Mosquitto client = Convert(cookie);
            if (client != null)
            {
                client.OnConnection(client, (ConnectionResult) rc);
            }
        }

        private static void OnDisconnectCallback(IntPtr mosquitto, IntPtr cookie, int rc)
        {
            Mosquitto client = Convert(cookie);
            if (client != null)
            {
                client.OnDisconnect(client, rc);
            }
        }

        private static void OnUnsubscribeCallback(IntPtr mosquitto, IntPtr cookie, int mid)
        {
            Mosquitto client = Convert(cookie);
            if (client != null)
            {
                client.OnUnsubscribe(client, mid);
            }
        }

        private static void OnPublishCallback(IntPtr mosquitto, IntPtr cookie, ushort mid)
        {
            Mosquitto client = Convert(cookie);
            if (client != null)
            {
                client.OnPublished(client, mid);
            }
        }

        private static void OnSubscribeCallback(IntPtr mosquitto, IntPtr cookie, int mid, int qosCount,byte[] grantedQos)
        {
            Mosquitto client = Convert(cookie);
            if (client != null)
            {
                client.OnSubscribe(client, mid, grantedQos);
            }
        }

        private static void OnMessageCallback(IntPtr mosquitto, IntPtr cookie, IntPtr message)
        {
            Mosquitto client = Convert(cookie);
            if (client != null)
            {
                Message m = ToMessage(message);
                try
                {
                    client.OnMessage(client, m);
                }
                catch
                {
                }
            }
        }

        private static void OnLogCallback(IntPtr mosquitto, IntPtr cookie,int level, string message)
        {
            Mosquitto client = Convert(cookie);
            if (client != null)
            {
                client.OnLog(client,(LogPriority) level, message);
            }
        }


        private static Message ToMessage(IntPtr message)
        {
            var m = (_Message) Marshal.PtrToStructure(message, typeof (_Message));
            var msg = new Message
                          {
                              Mid = m.mid,
                              Qos = (Qos)m.qos,
                              Retain = m.retain,
                              Topic = m.topic,
                              Payload = new byte[m.payloadlen]
                          };
            Marshal.Copy(m.payload, msg.Payload, 0, (int) m.payloadlen);
            return msg;
        }

        private static Mosquitto Convert(IntPtr cookie)
        {
            object client = GCHandle.FromIntPtr(cookie).Target;
            return (Mosquitto) client;
        }

        ~Mosquitto()
        {
            Dispose(false);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                Stop();
                if (disposing)
                {
                    // called from Dispose.
                    // dispose managed
                   _token.Dispose();
                   _task.Dispose();
                   
                }
                _this.Free();
                libmosquitto.mosquitto_destroy(_handle);
                if (--_libraryRefCounter == 0)
                {
                    // There are no unmanaged resources to release, but
                    // if we add them, they need to be released here.
                    libmosquitto.mosquitto_lib_cleanup();
                }
            }
            _disposed = true;
        }

        private void Resubscribe()
        {
            foreach(var topic in _subscriptions)
            {
                int mid = 0;
                Subscribe(ref mid, topic.Key, topic.Value);
            }
        }

        private void Stop()
        {
            if (_task.Status == TaskStatus.Running)
            {
                _token.Cancel();
                _task.Wait();
            }
        }

        private static readonly MessageCallback MessageCallback = OnMessageCallback;
        private static readonly SubscribeCallback SubscribeCallback = OnSubscribeCallback;
        private static readonly UnsubscribeCallback UnsubscribeCallback = OnUnsubscribeCallback;
        private static readonly PublishCallback PublishCallback = OnPublishCallback;
        private static readonly DisconnectCallback DisconnectCallback = OnDisconnectCallback;
        private static readonly ConnectCallback ConnectCallback = OnConnectCallback;
        private static readonly LogCallback LogCallback = OnLogCallback;
        private static int _libraryRefCounter;
        private readonly IntPtr _handle;
        private IObservable<Message> _observable ;
        private Task _task;
        private bool _disposed;
        private GCHandle _this;
        private CancellationTokenSource _token;
        private readonly Dictionary<string,Qos> _subscriptions = new Dictionary<string,Qos>();
        
    }
}