﻿using System;
using System.ComponentModel;
using System.Reflection;
using System.Runtime.InteropServices;

namespace mosquitto.net
{
    /// <summary>
    /// Mosquitto return codes
    /// </summary>
    public class ReturnCode
    {
        #region Code enum
        /// <summary>
        /// The enumeration of the return codes 
        /// </summary>
        public enum Code
        {
            /// <summary>
            /// Success
            /// </summary>
            [Description("Success")] Success = 0,
            /// <summary>
            /// No memory
            /// </summary>
            [Description("No memory")] NoMemory = 1,
            /// <summary>
            /// Protocol Error
            /// </summary>
            [Description("Protocol error")] ProtocolError = 2,
            /// <summary>
            /// Invalid Error
            /// </summary>
            [Description("Invalid")] InvalidError = 3,
            /// <summary>
            /// No Connection
            /// </summary>
            [Description("No connection")] NoConnection = 4,
            /// <summary>
            /// Connection is refused
            /// </summary>
            [Description("Connection Refused")] ConnectionRefused = 5,
            /// <summary>
            /// Not found
            /// </summary>
            [Description("Not found")] NotFound = 6,
            /// <summary>
            /// Connection lost
            /// </summary>
            [Description("Connection Lost")] ConnectionLost = 7,
            /// <summary>
            /// SSL Error
            /// </summary>
            [Description("SSL Error")] SslError = 8,
            /// <summary>
            /// Incorrect payload size
            /// </summary>
            [Description("Payload size")] PayloadSize = 9,
            /// <summary>
            /// No supported
            /// </summary>
            [Description("Not supported")] NotSupported = 10,
            /// <summary>
            /// Authorization error
            /// </summary>
            [Description("Authorization failed")] Authorization = 11,
            /// <summary>
            /// 
            /// </summary>
            [Description("ACL Denied")] AclDenied = 12,
            /// <summary>
            /// 
            /// </summary>
            [Description("Unknown error")] Unknown = 13,
            /// <summary>
            /// 
            /// </summary>
            [Description("Check the errno code")] ErrorNo = 14
        }

        #endregion

        public ReturnCode(int id)
        {
            Value = (Code) id;
        }

        public ReturnCode(Code id)
        {
            Value = id;
        }

        public Code Value
        {
            get
            {
                return _code;
            }  
            
            private set
            {
                _code = value;
                LastError = false;
                LastErrorDescription = "";
                if(value == Code.ErrorNo)
                {
                    LastError = true;
                    var error = libmosquitto.mosquitto_errno();
                    LastErrorDescription = new Win32Exception(error).Message;

                }
            }
        }

        public bool LastError { get; private set; }

        public string LastErrorDescription { get; set; }

        public static implicit operator ReturnCode(int code)
        {
            return new ReturnCode(code);
        }

        public static implicit operator ReturnCode(Code code)
        {
            return new ReturnCode(code);
        }

        public static implicit operator bool(ReturnCode code)
        {
            return code == Code.Success;
        }

        public static bool operator ==(ReturnCode code, int c)
        {
            if (code != null)
            {
                return code.Value == (Code) c;
            }
            return false;
        }

        public static bool operator ==(ReturnCode code, Code c)
        {
            if (code != null)
            {
                return code.Value == c;
            }
            return false;
        }

        public static bool operator !=(ReturnCode code, Code c)
        {
            if (code != null)
            {
                return code.Value != c;
            }
            return false;
        }

        public static bool operator !=(ReturnCode code, int c)
        {
            if (code != null)
            {
                return code.Value != (Code) c;
            }
            return false;
        }

        public override string ToString()
        {
            if(LastError)
            {
                return LastErrorDescription;
            }
            return Value.Description();
        }

        public override int GetHashCode()
        {
            return Value.GetHashCode();
        }

        public override bool Equals(object obj)
        {
            if (obj == null)
                return false;
            if (ReferenceEquals(this, obj))
                return true;
            var rhs = obj as ReturnCode;
            if (rhs == null)
                return false;
            return Value == rhs.Value;
        }

        private Code _code;
        
    }

    public static class Extensions
    {
        public static string Description(this ReturnCode.Code value)
        {
            FieldInfo fi = value.GetType().GetField(value.ToString());

            var attributes =
                (DescriptionAttribute[]) fi.GetCustomAttributes(
                    typeof (DescriptionAttribute),
                    false);

            if (attributes.Length > 0)
                return attributes[0].Description;
            return value.ToString();
        }
    }
}